"""DPI Restrictions as part of a UniFi network."""

from typing import Final

from ..models.dpi_restriction_app import DPIRestrictionApp
from .api import APIItems

APP_URL: Final = "/rest/dpiapp"  # List DPI App configuration


class DPIRestrictionApps(APIItems):
    """Represents DPI App configurations."""

    KEY = "_id"
    path = APP_URL
    item_cls = DPIRestrictionApp

    async def enable(self, app_id: str) -> list[dict]:
        """Enable DPI Restriction Group Apps."""
        app_url = f"{APP_URL}/{app_id}"
        data = {"enabled": True}
        return await self.controller.request("put", app_url, json=data)

    async def disable(self, app_id: str) -> list[dict]:
        """Disable DPI Restriction Group Apps."""
        app_url = f"{APP_URL}/{app_id}"
        data = {"enabled": False}
        return await self.controller.request("put", app_url, json=data)



"""Clients are devices on a UniFi network."""

from __future__ import annotations

from typing import Final

from ..models.client import Client
from .api import APIItems

URL_ALL: Final = "/rest/user"  # All known and configured clients


class ClientsAll(APIItems):
    """Represents all client network devices."""

    KEY = "mac"
    path = URL_ALL
    item_cls = Client



"""DPI Restrictions as part of a UniFi network."""

from .api import APIItem


class DPIRestrictionGroup(APIItem):
    """Represents a DPI Group configuration."""

    @property
    def id(self) -> str:
        """DPI group ID."""
        return self.raw["_id"]

    @property
    def attr_no_delete(self) -> bool:
        """Can be deleted."""
        return self.raw.get("attr_no_delete", False)

    @property
    def attr_hidden_id(self) -> str:
        """Attr hidden ID."""
        return self.raw.get("attr_hidden_id", "")

    @property
    def name(self) -> str:
        """DPI group name."""
        return self.raw["name"]

    @property
    def site_id(self) -> str:
        """Site ID."""
        return self.raw["site_id"]

    @property
    def dpiapp_ids(self) -> list[str]:
        """DPI app IDs belonging to group."""
        return self.raw.get("dpiapp_ids", [])



"""Event messages on state changes."""

#  https://demo.ui.com/manage/locales/en/eventStrings.json?v=5.4.11.2

from typing import Final, final

CONTROLLER_UPDATE_AVAILABLE: Final = "EVT_AD_Update_Available"

ACCESS_POINT_ADOPTED: Final = "EVT_AP_Adopted"
ACCESS_POINT_CONFIGURED: Final = "EVT_AP_Configured"
ACCESS_POINT_CONNECTED: Final = "EVT_AP_Connected"
ACCESS_POINT_DELETED: Final = "EVT_AP_Deleted"
ACCESS_POINT_LOST_CONTACT: Final = "EVT_AP_Lost_Contact"
ACCESS_POINT_RESTARTED: Final = "EVT_AP_Restarted"
ACCESS_POINT_RESTARTED_UNKNOWN: Final = "EVT_AP_RestartedUnknown"
ACCESS_POINT_UPGRADED: Final = "EVT_AP_Upgraded"

GATEWAY_ADOPTED: Final = "EVT_GW_Adopted"
GATEWAY_CONFIGURED: Final = "EVT_GW_Configured"
GATEWAY_CONNECTED: Final = "EVT_GW_Connected"
GATEWAY_DELETED: Final = "EVT_GW_Deleted"
GATEWAY_LOST_CONTACT: Final = "EVT_GW_Lost_Contact"
GATEWAY_RESTARTED: Final = "EVT_GW_Restarted"
GATEWAY_RESTARTED_UNKNOWN: Final = "EVT_GW_RestartedUnknown"
GATEWAY_UPGRADED: Final = "EVT_GW_Upgraded"

SWITCH_ADOPTED: Final = "EVT_SW_Adopted"
SWITCH_CONFIGURED: Final = "EVT_SW_Configured"
SWITCH_CONNECTED: Final = "EVT_SW_Connected"
SWITCH_DELETED: Final = "EVT_SW_Deleted"
SWITCH_LOST_CONTACT: Final = "EVT_SW_Lost_Contact"
SWITCH_OVERHEAT: Final = "EVT_SW_Overheat"
SWITCH_POE_OVERLOAD: Final = "EVT_SW_POE_Overload"
SWITCH_POE_DISCONNECT: Final = "EVT_SW_POE_Disconnect"
SWITCH_RESTARTED: Final = "EVT_SW_Restarted"
SWITCH_RESTARTED_UNKNOWN: Final = "EVT_SW_RestartedUnknown"
SWITCH_UPGRADED: Final = "EVT_SW_Upgraded"

WIRED_CLIENT_CONNECTED: Final = "EVT_LU_Connected"
WIRED_CLIENT_DISCONNECTED: Final = "EVT_LU_Disconnected"
WIRED_CLIENT_BLOCKED: Final = "EVT_LC_Blocked"
WIRED_CLIENT_UNBLOCKED: Final = "EVT_LC_Unblocked"
WIRELESS_CLIENT_CONNECTED: Final = "EVT_WU_Connected"
WIRELESS_CLIENT_DISCONNECTED: Final = "EVT_WU_Disconnected"
WIRELESS_CLIENT_BLOCKED: Final = "EVT_WC_Blocked"
WIRELESS_CLIENT_UNBLOCKED: Final = "EVT_WC_Unblocked"
WIRELESS_CLIENT_ROAM: Final = "EVT_WU_Roam"
WIRELESS_CLIENT_ROAMRADIO: Final = "EVT_WU_RoamRadio"
WIRELESS_GUEST_CONNECTED: Final = "EVT_WG_Connected"
WIRELESS_GUEST_DISCONNECTED: Final = "EVT_WG_Disconnected"
WIRELESS_GUEST_ROAM: Final = "EVT_WG_Roam"
WIRELESS_GUEST_ROAMRADIO: Final = "EVT_WG_RoamRadio"

CLIENT_EVENTS: Final = (
    WIRED_CLIENT_CONNECTED,
    WIRED_CLIENT_DISCONNECTED,
    WIRED_CLIENT_BLOCKED,
    WIRED_CLIENT_UNBLOCKED,
    WIRELESS_CLIENT_CONNECTED,
    WIRELESS_CLIENT_DISCONNECTED,
    WIRELESS_CLIENT_BLOCKED,
    WIRELESS_CLIENT_UNBLOCKED,
    WIRELESS_CLIENT_ROAM,
    WIRELESS_CLIENT_ROAMRADIO,
    WIRELESS_GUEST_CONNECTED,
    WIRELESS_GUEST_DISCONNECTED,
    WIRELESS_GUEST_ROAM,
    WIRELESS_GUEST_ROAMRADIO,
)
DEVICE_EVENTS: Final = (
    ACCESS_POINT_ADOPTED,
    ACCESS_POINT_CONFIGURED,
    ACCESS_POINT_CONNECTED,
    ACCESS_POINT_DELETED,
    ACCESS_POINT_LOST_CONTACT,
    ACCESS_POINT_RESTARTED,
    ACCESS_POINT_RESTARTED_UNKNOWN,
    ACCESS_POINT_UPGRADED,
    GATEWAY_ADOPTED,
    GATEWAY_CONFIGURED,
    GATEWAY_CONNECTED,
    GATEWAY_DELETED,
    GATEWAY_LOST_CONTACT,
    GATEWAY_RESTARTED,
    GATEWAY_RESTARTED_UNKNOWN,
    GATEWAY_UPGRADED,
    SWITCH_ADOPTED,
    SWITCH_CONFIGURED,
    SWITCH_CONNECTED,
    SWITCH_DELETED,
    SWITCH_LOST_CONTACT,
    SWITCH_OVERHEAT,
    SWITCH_POE_OVERLOAD,
    SWITCH_POE_DISCONNECT,
    SWITCH_RESTARTED,
    SWITCH_RESTARTED_UNKNOWN,
    SWITCH_UPGRADED,
)


@final
class Event:
    """UniFi event."""

    def __init__(self, raw: dict) -> None:
        """Initialize event."""
        self.raw = raw

    @property
    def datetime(self) -> str:
        """Datetime of event '2020-03-01T15:35:08Z'."""
        return self.raw["datetime"]

    @property
    def event(self) -> str:
        """Event key e.g. 'EVT_WU_Disconnected'."""
        return self.raw["key"]

    @property
    def msg(self) -> str:
        """Message 'User[00:00:00:00:00:01] disconnected from "Access point" (1h 27m connected, 58.97M bytes, last AP[00:11:22:33:44:55])'."""
        return self.raw["msg"]

    @property
    def time(self) -> int:
        """Time of event 1583076908000."""
        return self.raw["time"]

    @property
    def mac(self) -> str:
        """MAC of client or device."""
        if self.client:
            return self.client
        if self.device:
            return self.device
        return ""

    @property
    def ap(self) -> str:
        """Access point connected to."""
        return self.raw.get("ap", "")

    @property
    def bytes(self) -> int:
        """Bytes of data consumed."""
        return self.raw.get("bytes", 0)

    @property
    def channel(self) -> int:
        """Wi-Fi channel."""
        return self.raw.get("channel", 0)

    @property
    def client(self) -> str:
        """MAC address of client."""
        return (
            self.raw.get("user")
            or self.raw.get("client")
            or self.raw.get("guest")
            or ""
        )

    @property
    def device(self) -> str:
        """MAC address of device."""
        return self.raw.get("ap") or self.raw.get("gw") or self.raw.get("sw") or ""

    @property
    def duration(self) -> int:
        """Duration."""
        return self.raw.get("duration", 0)

    @property
    def hostname(self) -> str:
        """Nice name."""
        return self.raw.get("hostname", "")

    @property
    def radio(self) -> str:
        """Radio."""
        return self.raw.get("radio", "")

    @property
    def subsystem(self) -> str:
        """Subsystem like 'lan' or 'wlan'."""
        return self.raw.get("subsystem", "")

    @property
    def site_id(self) -> str:
        """Site ID."""
        return self.raw.get("site_id", "")

    @property
    def ssid(self) -> str:
        """SSID."""
        return self.raw.get("ssid", "")

    @property
    def version_from(self) -> str:
        """Version from."""
        return self.raw.get("version_from", "")

    @property
    def version_to(self) -> str:
        """Version to."""
        return self.raw.get("version_to", "")



"""WLANs as part of a UniFi network."""

from .api import APIItem


class Wlan(APIItem):
    """Represent a WLAN configuration."""

    @property
    def id(self) -> str:
        """ID of WLAN."""
        return self.raw["_id"]

    @property
    def bc_filter_enabled(self) -> bool:
        """Is BC filter enabled."""
        return self.raw.get("bc_filter_enabled", False)

    @property
    def bc_filter_list(self) -> list:
        """List of BC filters."""
        return self.raw["bc_filter_list"]

    @property
    def dtim_mode(self) -> str:
        """DTIM mode."""
        return self.raw["dtim_mode"]

    @property
    def dtim_na(self) -> int:
        """DTIM NA."""
        return self.raw["dtim_na"]

    @property
    def dtim_ng(self) -> int:
        """DTIM NG."""
        return self.raw["dtim_ng"]

    @property
    def enabled(self) -> bool:
        """Is WLAN enabled."""
        return self.raw["enabled"]

    @property
    def group_rekey(self) -> int:
        """Group rekey."""
        return self.raw["group_rekey"]

    @property
    def is_guest(self) -> bool:
        """Is WLAN a guest network."""
        return self.raw.get("is_guest", False)

    @property
    def mac_filter_enabled(self) -> bool:
        """Is MAC filtering enabled."""
        return self.raw.get("mac_filter_enabled", False)

    @property
    def mac_filter_list(self) -> list:
        """List of MAC filters."""
        return self.raw["mac_filter_list"]

    @property
    def mac_filter_policy(self) -> str:
        """MAC filter policy."""
        return self.raw["mac_filter_policy"]

    @property
    def minrate_na_advertising_rates(self) -> bool:
        """Minrate NA advertising rates."""
        return self.raw["minrate_na_advertising_rates"]

    @property
    def minrate_na_beacon_rate_kbps(self) -> int:
        """Minrate NA beacon rate kbps."""
        return self.raw["minrate_na_beacon_rate_kbps"]

    @property
    def minrate_na_data_rate_kbps(self) -> int:
        """Minrate NA data rate kbps."""
        return self.raw["minrate_na_data_rate_kbps"]

    @property
    def minrate_na_enabled(self) -> bool:
        """Is minrate NA enabled."""
        return self.raw.get("minrate_na_enabled", False)

    @property
    def minrate_na_mgmt_rate_kbps(self) -> int:
        """Minrate NA management rate kbps."""
        return self.raw["minrate_na_mgmt_rate_kbps"]

    @property
    def minrate_ng_advertising_rates(self) -> bool:
        """Minrate NG advertising rates."""
        return self.raw["minrate_ng_advertising_rates"]

    @property
    def minrate_ng_beacon_rate_kbps(self) -> int:
        """Minrate NG beacon rate kbps."""
        return self.raw["minrate_ng_beacon_rate_kbps"]

    @property
    def minrate_ng_cck_rates_enabled(self) -> bool:
        """Is minrate NG CCK rates enabled."""
        return self.raw.get("minrate_ng_cck_rates_enabled", False)

    @property
    def minrate_ng_data_rate_kbps(self) -> int:
        """Minrate NG data rate kbps."""
        return self.raw["minrate_ng_data_rate_kbps"]

    @property
    def minrate_ng_enabled(self) -> bool:
        """Is minrate NG enabled."""
        return self.raw.get("minrate_ng_enabled", False)

    @property
    def minrate_ng_mgmt_rate_kbps(self) -> int:
        """Minrate NG management rate kbps."""
        return self.raw["minrate_ng_mgmt_rate_kbps"]

    @property
    def name(self) -> str:
        """WLAN name."""
        return self.raw["name"]

    @property
    def name_combine_enabled(self) -> bool:
        """If 2.5 and 5 GHz SSIDs should be combined."""
        return self.raw.get("name_combine_enabled", True)

    @property
    def name_combine_suffix(self) -> str:
        """Suffix for 2.4GHz SSID if name is not combined."""
        return self.raw.get("name_combine_suffix", "")

    @property
    def no2ghz_oui(self) -> bool:
        """No 2GHz OUI."""
        return self.raw["no2ghz_oui"]

    @property
    def schedule(self) -> list:
        """Schedule list."""
        return self.raw["schedule"]

    @property
    def security(self) -> str:
        """Security."""
        return self.raw["security"]

    @property
    def site_id(self) -> str:
        """WLAN site ID."""
        return self.raw["site_id"]

    @property
    def usergroup_id(self) -> str:
        """WLAN user group ID."""
        return self.raw["usergroup_id"]

    @property
    def wep_idx(self) -> int:
        """WEP index."""
        return self.raw["wep_idx"]

    @property
    def wlangroup_id(self) -> str:
        """WLAN group ID."""
        return self.raw["wlangroup_id"]

    @property
    def wpa_enc(self) -> str:
        """WPA encryption."""
        return self.raw["wpa_enc"]

    @property
    def wpa_mode(self) -> str:
        """WPA mode."""
        return self.raw["wpa_mode"]

    @property
    def x_iapp_key(self) -> str:
        """X iapp key."""
        return self.raw["x_iapp_key"]

    @property
    def x_passphrase(self) -> str:
        """X passphrase."""
        return self.raw["x_passphrase"]



"""Aiounifi errors."""


class AiounifiException(Exception):
    """Base error for aiounifi."""


class RequestError(AiounifiException):
    """Unable to fulfill request.

    Raised when host or API cannot be reached.
    """


class ResponseError(AiounifiException):
    """Invalid response."""


class Unauthorized(AiounifiException):
    """Username is not authorized."""


class LoginRequired(AiounifiException):
    """User is logged out."""


class NoPermission(AiounifiException):
    """Users permissions are read only."""


class ServiceUnavailable(RequestError):
    """Service is unavailable.

    Common error if controller is restarting and behind a proxy.
    """


class BadGateway(RequestError):
    """Invalid response from the upstream server."""


class TwoFaTokenRequired(AiounifiException):
    """2 factor authentication token required."""


ERRORS = {
    "api.err.LoginRequired": LoginRequired,
    "api.err.Invalid": Unauthorized,
    "api.err.NoPermission": NoPermission,
    "api.err.Ubic2faTokenRequired": TwoFaTokenRequired,
}


def raise_error(error) -> None:
    """Raise error."""
    type = error
    cls = ERRORS.get(type, AiounifiException)
    raise cls("{}".format(type))



"""Clients are devices on a UniFi network."""

from __future__ import annotations

from typing import Final

from ..models.client import Client
from .api import APIItems

URL: Final = "/stat/sta"  # Active clients

URL_CLIENT_STATE_MANAGER: Final = "/cmd/stamgr"


class Clients(APIItems):
    """Represents client network devices."""

    KEY = "mac"
    path = URL
    item_cls = Client

    async def block(self, mac: str) -> list[dict]:
        """Block client from controller."""
        data = {"mac": mac, "cmd": "block-sta"}
        return await self.controller.request(
            "post", URL_CLIENT_STATE_MANAGER, json=data
        )

    async def unblock(self, mac: str) -> list[dict]:
        """Unblock client from controller."""
        data = {"mac": mac, "cmd": "unblock-sta"}
        return await self.controller.request(
            "post", URL_CLIENT_STATE_MANAGER, json=data
        )

    async def reconnect(self, mac: str) -> list[dict]:
        """Force a wireless client to reconnect to the network."""
        data = {"mac": mac, "cmd": "kick-sta"}
        return await self.controller.request(
            "post", URL_CLIENT_STATE_MANAGER, json=data
        )

    async def remove_clients(self, macs: list[str]) -> list[dict]:
        """Make controller forget provided clients."""
        data = {"macs": macs, "cmd": "forget-sta"}
        return await self.controller.request(
            "post", URL_CLIENT_STATE_MANAGER, json=data
        )



"""UniFi devices are network infrastructure.

Access points, Gateways, Switches.
"""

from __future__ import annotations

from typing import Final

from ..models.device import Device
from .api import APIItems

URL: Final = "/stat/device"

URL_DEVICE_MANAGER: Final = "/cmd/devmgr"


class Devices(APIItems):
    """Represents network devices."""

    KEY = "mac"
    path = URL
    item_cls = Device

    async def upgrade(self, mac: str) -> list[dict]:
        """Upgrade network device."""
        data = {"mac": mac, "cmd": "upgrade"}
        return await self.controller.request("post", URL_DEVICE_MANAGER, json=data)



"""UniFi devices are network infrastructure.

Access points, Gateways, Switches.
"""

from __future__ import annotations

from collections.abc import Awaitable, Callable, Iterator, ValuesView
import logging
from typing import Final

from ..events import Event as UniFiEvent
from .api import APIItem

LOGGER = logging.getLogger(__name__)

URL: Final = "/stat/device"


class Device(APIItem):
    """Represents a network device."""

    def __init__(
        self,
        raw: dict,
        request: Callable[..., Awaitable[list[dict]]],
    ) -> None:
        """Initialize device."""
        super().__init__(raw, request)
        self.ports = Ports(raw.get("port_table", []))
        self.outlets = Outlets(raw.get("outlet_table", []))

    def update(
        self,
        raw: dict | None = None,
        event: UniFiEvent | None = None,
    ) -> None:
        """Refresh data."""
        if raw:
            self.ports.update(raw.get("port_table", []))
            self.outlets.update(raw.get("outlet_table", []))
        super().update(raw, event)

    @property
    def board_revision(self) -> int:
        """Board revision of device."""
        return self.raw["board_rev"]

    @property
    def considered_lost_at(self) -> int:
        """When device is considered lost."""
        return self.raw["considered_lost_at"]

    @property
    def disabled(self) -> bool:
        """Is device disabled."""
        return self.raw.get("disabled", False)

    @property
    def downlink_table(self) -> list[dict]:
        """All devices with device as uplink."""
        return self.raw.get("downlink_table", [])

    @property
    def id(self) -> str:
        """ID of device."""
        return self.raw["device_id"]

    @property
    def ip(self) -> str:
        """IP of device."""
        return self.raw["ip"]

    @property
    def fan_level(self) -> int | None:
        """Fan level of device."""
        return self.raw.get("fan_level")

    @property
    def has_fan(self) -> bool:
        """Do device have a fan."""
        return self.raw.get("has_fan", False)

    @property
    def last_seen(self) -> int | None:
        """When was device last seen."""
        return self.raw.get("last_seen")

    @property
    def lldp_table(self) -> list[dict]:
        """All clients and devices directly attached to device."""
        return self.raw.get("lldp_table", [])

    @property
    def mac(self) -> str:
        """MAC address of device."""
        return self.raw["mac"]

    @property
    def model(self) -> str:
        """Model of device."""
        return self.raw["model"]

    @property
    def name(self) -> str:
        """Name of device."""
        return self.raw.get("name", "")

    @property
    def next_heartbeat_at(self) -> int | None:
        """Next heart beat full UNIX time."""
        return self.raw.get("next_heartbeat_at")

    @property
    def next_interval(self) -> int:
        """Next heart beat in seconds."""
        return self.raw.get("next_interval", 30)

    @property
    def overheating(self) -> bool:
        """Is device overheating."""
        return self.raw.get("overheating", False)

    @property
    def outlet_overrides(self) -> list:
        """Overridden outlet configuration."""
        return self.raw.get("outlet_overrides", [])

    @property
    def outlet_table(self) -> list:
        """List of outlets."""
        return self.raw.get("outlet_table", [])

    @property
    def port_overrides(self) -> list:
        """Overridden port configuration."""
        return self.raw.get("port_overrides", [])

    @property
    def port_table(self) -> list:
        """List of ports and data."""
        return self.raw.get("port_table", [])

    @property
    def state(self) -> int:
        """State of device."""
        return self.raw["state"]

    @property
    def sys_stats(self) -> dict:
        """Output from top."""
        return self.raw["sys_stats"]

    @property
    def type(self) -> str:
        """Type of device."""
        return self.raw["type"]

    @property
    def version(self) -> str:
        """Firmware version."""
        return self.raw["version"]

    @property
    def upgradable(self) -> bool:
        """Is a new firmware available."""
        return self.raw.get("upgradable", False)

    @property
    def upgrade_to_firmware(self) -> str:
        """Firmware version to update to."""
        return self.raw.get("upgrade_to_firmware", "")

    @property
    def uplink(self) -> dict[str, bool | int | list[str] | str]:
        """Information about uplink."""
        return self.raw["uplink"]

    @property
    def uplink_depth(self) -> int | None:
        """Hops to gateway."""
        return self.raw.get("uplink_depth")

    @property
    def user_num_sta(self) -> int:
        """Amount of connected clients."""
        return self.raw["user-num_sta"]

    @property
    def wlan_overrides(self) -> list:
        """Wlan configuration override."""
        return self.raw.get("wlan_overrides", [])

    async def set_outlet_relay_state(self, outlet_idx: int, state: bool) -> list[dict]:
        """Set outlet relay state.

        True:  outlet power output on.
        False: outlet power output off.
        """
        LOGGER.debug("Override outlet %d with relay_state %s", outlet_idx, str(state))

        existing_override = False
        for outlet_override in self.outlet_overrides:
            if outlet_idx == outlet_override["index"]:
                outlet_override["relay_state"] = state
                existing_override = True
                break

        if not existing_override:
            self.outlet_overrides.append(
                {
                    "index": outlet_idx,
                    "name": self.outlets[outlet_idx].name,
                    "relay_state": state,
                }
            )
        url = f"/rest/device/{self.id}"
        data = {"outlet_overrides": self.outlet_overrides}

        return await self._request("put", url, json=data)

    async def set_outlet_cycle_enabled(
        self, outlet_idx: int, state: bool
    ) -> list[dict]:
        """Set outlet cycle_enabled flag.

        True:  UniFi Network will power cycle this outlet if the internet goes down.
        False: UniFi Network will not power cycle this outlet if the internet goes down.
        """
        LOGGER.debug("Override outlet %d with cycle_enabled %s", outlet_idx, str(state))

        existing_override = False
        for outlet_override in self.outlet_overrides:
            if outlet_idx == outlet_override["index"]:
                outlet_override["cycle_enabled"] = state
                existing_override = True
                break

        if not existing_override:
            self.outlet_overrides.append(
                {
                    "index": outlet_idx,
                    "name": self.outlets[outlet_idx].name,
                    "cycle_enabled": state,
                }
            )
        url = f"/rest/device/{self.id}"
        data = {"outlet_overrides": self.outlet_overrides}

        return await self._request("put", url, json=data)

    async def set_port_poe_mode(self, port_idx: int, mode: str) -> list[dict]:
        """Set port poe mode.

        Auto, 24v, passthrough, off.
        Make sure to not overwrite any existing configs.
        """
        LOGGER.debug("Override port %d with mode %s", port_idx, mode)

        existing_override = False
        for port_override in self.port_overrides:
            if port_idx == port_override["port_idx"]:
                port_override["poe_mode"] = mode
                existing_override = True
                break

        if not existing_override:
            self.port_overrides.append(
                {
                    "port_idx": port_idx,
                    "portconf_id": self.ports[port_idx].portconf_id,
                    "poe_mode": mode,
                }
            )

        url = f"/rest/device/{self.id}"
        data = {"port_overrides": self.port_overrides}

        return await self._request("put", url, json=data)

    def __repr__(self) -> str:
        """Return the representation."""
        return f"<Device {self.name}: {self.mac}>"


class Port:
    """Represents a network port."""

    def __init__(self, raw: dict) -> None:
        """Initialize port."""
        self.raw = raw

    @property
    def ifname(self) -> str:
        """Port name used by USG."""
        return self.raw.get("ifname", "")

    @property
    def media(self) -> str:
        """Media port is connected to."""
        return self.raw.get("media", "")

    @property
    def name(self) -> str:
        """Port name."""
        return self.raw["name"]

    @property
    def port_idx(self) -> int | None:
        """Port index."""
        return self.raw.get("port_idx")

    @property
    def poe_class(self) -> str:
        """Port POE class."""
        return self.raw.get("poe_class", "")

    @property
    def poe_enable(self) -> bool | None:
        """Is POE supported/requested by client."""
        return self.raw.get("poe_enable")

    @property
    def poe_mode(self) -> str:
        """Is POE auto, pasv24, passthrough, off or None."""
        return self.raw.get("poe_mode", "")

    @property
    def poe_power(self) -> str:
        """POE power usage."""
        return self.raw.get("poe_power", "")

    @property
    def poe_voltage(self) -> str:
        """POE voltage usage."""
        return self.raw.get("poe_voltage", "")

    @property
    def portconf_id(self) -> str:
        """Port configuration ID."""
        return self.raw.get("portconf_id", "")

    @property
    def port_poe(self) -> bool:
        """Is POE used."""
        return self.raw.get("port_poe") is True

    @property
    def up(self) -> str:
        """Is port up."""
        return self.raw.get("up", "")

    def __repr__(self) -> str:
        """Return the representation."""
        return f"<{self.name}: Poe {self.poe_enable}>"


class Ports:
    """Represents ports on a device."""

    def __init__(self, raw: list[dict]) -> None:
        """Initialize port manager."""
        self.ports: dict[int | str, Port] = {}
        for raw_port in raw:
            port = Port(raw_port)

            if (index := port.port_idx) is not None:
                self.ports[index] = port
            elif ifname := port.ifname:
                self.ports[ifname] = port

    def update(self, raw: list[dict]) -> None:
        """Update ports."""
        for raw_port in raw:
            index = None

            if "port_idx" in raw_port:
                index = raw_port["port_idx"]

            elif "ifname" in raw_port:
                index = raw_port["ifname"]

            if index in self.ports:
                self.ports[index].raw = raw_port

    def values(self) -> ValuesView[Port]:
        """Return ports."""
        return self.ports.values()

    def __getitem__(self, obj_id: int) -> Port:
        """Get specific port based on key."""
        return self.ports[obj_id]

    def __iter__(self) -> Iterator[int | str]:
        """Iterate over ports."""
        return iter(self.ports)


class Outlet:
    """Represents an outlet."""

    def __init__(self, raw: dict) -> None:
        """Initialize outlet."""
        self.raw = raw

    @property
    def name(self) -> str:
        """Name of outlet."""
        return self.raw["name"]

    @property
    def index(self) -> int:
        """Outlet index."""
        return self.raw["index"]

    @property
    def has_relay(self) -> bool:
        """Is the outlet controllable."""
        return self.raw.get("has_relay", False)

    @property
    def has_metering(self) -> bool:
        """Is metering supported."""
        return self.raw["has_metering"]

    @property
    def relay_state(self) -> bool:
        """Is outlet power on."""
        return self.raw["relay_state"]

    @property
    def cycle_enabled(self) -> bool:
        """Modem Power Cycle."""
        return self.raw.get("cycle_enabled", False)


class Outlets:
    """Represents outlets on a device."""

    def __init__(self, raw: list[dict]) -> None:
        """Initialize outlet manager."""
        self.outlets: dict[int, Outlet] = {}
        for raw_outlet in raw:
            outlet = Outlet(raw_outlet)
            self.outlets[outlet.index] = outlet

    def update(self, raw: list[dict]) -> None:
        """Update outlets."""
        for raw_outlet in raw:
            index = raw_outlet["index"]
            self.outlets[index].raw = raw_outlet

    def values(self) -> ValuesView[Outlet]:
        """Return outlets."""
        return self.outlets.values()

    def __getitem__(self, obj_id: int) -> Outlet:
        """Get specific outlet based on key."""
        return self.outlets[obj_id]

    def __iter__(self) -> Iterator[int]:
        """Iterate over outlets."""
        return iter(self.outlets)



"""API management class and base class for the different end points."""

from __future__ import annotations

from collections.abc import Callable, ItemsView, Iterator, ValuesView
import logging
from typing import Any, Final, final

from ..events import Event as UniFiEvent

SubscriptionType = Callable[[str, str], None]

LOGGER = logging.getLogger(__name__)

SOURCE_DATA: Final = "data"
SOURCE_EVENT: Final = "event"


class APIItems:
    """Base class for a map of API Items."""

    KEY = ""
    path = ""
    item_cls: Any

    def __init__(self, controller) -> None:
        """Initialize API items."""
        self.controller = controller
        self._items: dict[int | str, Any] = {}
        self._subscribers: list[SubscriptionType] = []

    @final
    async def update(self) -> None:
        """Refresh data."""
        raw = await self.controller.request("get", self.path)
        self.process_raw(raw)

    @final
    def process_raw(self, raw: list[dict]) -> set:
        """Process data."""
        new_items = set()

        for raw_item in raw:
            key = raw_item[self.KEY]

            if (obj := self._items.get(key)) is not None:
                obj.update(raw=raw_item)
                continue

            self._items[key] = self.item_cls(raw_item, self.controller.request)
            new_items.add(key)

            for callback in self._subscribers:
                callback("added", key)

        return new_items

    @final
    def process_event(self, events: list[UniFiEvent]) -> set:
        """Process event."""
        new_items = set()

        for event in events:

            if (obj := self._items.get(event.mac)) is not None:
                obj.update(event=event)
                new_items.add(event.mac)

        return new_items

    @final
    def remove(self, raw: list) -> set:
        """Remove list of items."""
        removed_items = set()

        for raw_item in raw:
            key = raw_item[self.KEY]

            if key in self._items:
                item = self._items.pop(key)
                item.clear_callbacks()
                removed_items.add(key)

        return removed_items

    def subscribe(self, callback: SubscriptionType) -> Callable:
        """Subscribe to added events.

        "callback" - callback function to call when an event emits.
        Return function to unsubscribe.
        """
        self._subscribers.append(callback)

        def unsubscribe():
            self._subscribers.remove(callback)

        return unsubscribe

    @final
    def items(self) -> ItemsView[int | str, Any]:
        """Return item values."""
        return self._items.items()

    @final
    def values(self) -> ValuesView[Any]:
        """Return item values."""
        return self._items.values()

    @final
    def get(
        self,
        obj_id: int | str,
        default: Any | None = None,
    ) -> Any | None:
        """Get item value based on key, return default if no match."""
        return self._items.get(obj_id, default)

    @final
    def __contains__(self, obj_id: int | str) -> bool:
        """Validate membership of item ID."""
        return obj_id in self._items

    @final
    def __getitem__(self, obj_id: int | str) -> Any:
        """Get item value based on key."""
        return self._items[obj_id]

    @final
    def __iter__(self) -> Iterator[int | str]:
        """Allow iterate over items."""
        return iter(self._items)



"""WLANs as part of a UniFi network."""

from typing import Final

from ..models.wlan import Wlan
from .api import APIItems

URL: Final = "/rest/wlanconf"  # List WLAN configuration


class Wlans(APIItems):
    """Represents WLAN configurations."""

    KEY = "name"
    path = URL
    item_cls = Wlan

    async def enable(self, wlan: Wlan) -> list[dict]:
        """Block client from controller."""
        wlan_url = f"{URL}/{wlan.id}"
        data = {"enabled": True}
        return await self.controller.request("put", wlan_url, json=data)

    async def disable(self, wlan: Wlan) -> list[dict]:
        """Unblock client from controller."""
        wlan_url = f"{URL}/{wlan.id}"
        data = {"enabled": False}
        return await self.controller.request("put", wlan_url, json=data)



"""Clients are devices on a UniFi network."""

from __future__ import annotations

from .api import APIItem


class Client(APIItem):
    """Represents a client network device."""

    @property
    def access_point_mac(self) -> str:
        """MAC address of access point."""
        return self.raw.get("ap_mac", "")

    @property
    def association_time(self) -> int:
        """When was client associated with controller."""
        return self.raw.get("assoc_time", 0)

    @property
    def blocked(self) -> bool:
        """Is client blocked."""
        return self.raw.get("blocked", False)

    @property
    def device_name(self) -> str:
        """Device name of client."""
        return self.raw.get("device_name", "")

    @property
    def essid(self) -> str:
        """ESSID client is connected to."""
        return self.raw.get("essid", "")

    @property
    def firmware_version(self) -> str:
        """Firmware version of client."""
        return self.raw.get("fw_version", "")

    @property
    def first_seen(self) -> int:
        """When was client first seen."""
        return self.raw.get("first_seen", 0)

    @property
    def fixed_ip(self) -> str:
        """List IP if fixed IP is configured."""
        return self.raw.get("fixed_ip", "")

    @property
    def hostname(self) -> str:
        """Hostname of client."""
        return self.raw.get("hostname", "")

    @property
    def idle_time(self) -> int:
        """Idle time of client."""
        return self.raw.get("idletime", 0)

    @property
    def ip(self) -> str:
        """IP of client."""
        return self.raw.get("ip", "")

    @property
    def is_guest(self) -> bool:
        """Is client guest."""
        return self.raw.get("is_guest", False)

    @property
    def is_wired(self) -> bool:
        """Is client wired."""
        return self.raw.get("is_wired", False)

    @property
    def last_seen(self) -> int:
        """When was client last seen."""
        return self.raw.get("last_seen", 0)

    @property
    def last_seen_by_access_point(self) -> int:
        """When was client last seen by access point."""
        return self.raw.get("_last_seen_by_uap", 0)

    @property
    def last_seen_by_gateway(self) -> int:
        """When was client last seen by gateway."""
        return self.raw.get("_last_seen_by_ugw", 0)

    @property
    def last_seen_by_switch(self) -> int:
        """When was client last seen by network switch."""
        return self.raw.get("_last_seen_by_usw", 0)

    @property
    def latest_association_time(self) -> int:
        """When was client last associated with controller."""
        return self.raw.get("latest_assoc_time", 0)

    @property
    def mac(self) -> str:
        """MAC address of client."""
        return self.raw.get("mac", "")

    @property
    def name(self) -> str:
        """Name of client."""
        return self.raw.get("name", "")

    @property
    def oui(self) -> str:
        """Vendor string for client MAC."""
        return self.raw.get("oui", "")

    @property
    def powersave_enabled(self) -> bool | None:
        """Powersave functionality enabled for wireless client."""
        return self.raw.get("powersave_enabled")

    @property
    def site_id(self) -> str:
        """Site ID client belongs to."""
        return self.raw.get("site_id", "")

    @property
    def switch_depth(self) -> int | None:
        """How many layers of switches client is in."""
        return self.raw.get("sw_depth")

    @property
    def switch_mac(self) -> str:
        """MAC for switch client is connected to."""
        return self.raw.get("sw_mac", "")

    @property
    def switch_port(self) -> int | None:
        """Switch port client is connected to."""
        return self.raw.get("sw_port")

    @property
    def rx_bytes(self) -> int:
        """Bytes received over wireless connection."""
        return self.raw.get("rx_bytes", 0)

    @property
    def rx_bytes_r(self) -> int:
        """Bytes recently received over wireless connection."""
        return self.raw.get("rx_bytes-r", 0)

    @property
    def tx_bytes(self) -> int:
        """Bytes transferred over wireless connection."""
        return self.raw.get("tx_bytes", 0)

    @property
    def tx_bytes_r(self) -> int:
        """Bytes recently transferred over wireless connection."""
        return self.raw.get("tx_bytes-r", 0)

    @property
    def uptime(self) -> int:
        """Uptime of client."""
        return self.raw.get("uptime", 0)

    @property
    def uptime_by_access_point(self) -> int:
        """Uptime of client observed by access point."""
        return self.raw.get("_uptime_by_uap", 0)

    @property
    def uptime_by_gateway(self) -> int:
        """Uptime of client observed by gateway."""
        return self.raw.get("_uptime_by_ugw", 0)

    @property
    def uptime_by_switch(self) -> int:
        """Uptime of client observed by network switch."""
        return self.raw.get("_uptime_by_usw", 0)

    @property
    def wired_rate_mbps(self) -> int:
        """Wired rate in Mbps."""
        return self.raw.get("wired_rate_mbps", 0)

    @property
    def wired_rx_bytes(self) -> int:
        """Bytes received over wired connection."""
        return self.raw.get("wired-rx_bytes", 0)

    @property
    def wired_rx_bytes_r(self) -> int:
        """Bytes recently received over wired connection."""
        return self.raw.get("wired-rx_bytes-r", 0)

    @property
    def wired_tx_bytes(self) -> int:
        """Bytes transferred over wired connection."""
        return self.raw.get("wired-tx_bytes", 0)

    @property
    def wired_tx_bytes_r(self) -> int:
        """Bytes recently transferred over wired connection."""
        return self.raw.get("wired-tx_bytes-r", 0)



"""DPI Restrictions as part of a UniFi network."""

from .api import APIItem


class DPIRestrictionApp(APIItem):
    """Represents a DPI App configuration."""

    @property
    def id(self) -> str:
        """DPI app ID."""
        return self.raw["_id"]

    @property
    def apps(self) -> list:
        """List of apps."""
        return self.raw["apps"]

    @property
    def blocked(self) -> bool:
        """Is blocked."""
        return self.raw["blocked"]

    @property
    def cats(self) -> list:
        """Categories."""
        return self.raw["cats"]

    @property
    def enabled(self) -> bool:
        """Is enabled."""
        return self.raw["enabled"]

    @property
    def log(self) -> bool:
        """Is logging enabled."""
        return self.raw["log"]

    @property
    def site_id(self) -> str:
        """Site ID."""
        return self.raw["site_id"]



"""DPI Restrictions as part of a UniFi network."""

from typing import Final

from ..models.dpi_restriction_group import DPIRestrictionGroup
from .api import APIItems

GROUP_URL: Final = "/rest/dpigroup"  # List DPI Group configuration


class DPIRestrictionGroups(APIItems):
    """Represents DPI Group configurations."""

    KEY = "_id"
    path = GROUP_URL
    item_cls = DPIRestrictionGroup
